@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2014  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -22,30 +22,23 @@
 
 #define uninitialized_var(x) x = x
 
+/*
+ * Set a maximum to prevent sending too much data to the kernel when hashing
+ * the password (or any other crypto operations involving the password).
+ * This value is not tied to IEEE or any RFC's, just chosen to be long enough
+ */
+#define IWD_MAX_PASSWORD_LEN	2048
+
 struct l_genl;
 struct l_genl_family;
 
 const struct l_settings *iwd_get_config(void);
 struct l_genl *iwd_get_genl(void);
 
-bool netdev_init(void);
-void netdev_exit(void);
-void netdev_set_nl80211(struct l_genl_family *nl80211);
 void netdev_shutdown(void);
 
-bool manager_init(struct l_genl_family *in,
-			const char *if_whitelist, const char *if_blacklist);
-void manager_exit(void);
+const char *iwd_get_iface_whitelist(void);
+const char *iwd_get_iface_blacklist(void);
 
-struct iwd_module_desc {
-	const char *name;
-	int (*init)(void);
-	void (*exit)(void);
-	bool active;
-} __attribute__((aligned(8)));
-
-#define IWD_MODULE(name, init, exit)					\
-	static struct iwd_module_desc __iwd_module_ ## name		\
-		__attribute__((used, section("__iwd_module"), aligned(8))) = {\
-			#name, init, exit				\
-		};
+const char *iwd_get_phy_whitelist(void);
+const char *iwd_get_phy_blacklist(void);
